/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

.toggle {
    display: block;
    clear: both;
}

.toggle:after {
    content: "Show Code [+]";
}

.toggle.open:before {
    content: "Hide Code [-]";
}
.toggle.open:after {
    content: "";
}
